# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.


from cryptography.__about__ import (
    __author__,
    __copyright__,
    __version__,
)


__all__ = [
    "__version__",
    "__author__",
    "__copyright__",
]
